package Main;

import java.io.*;

import ErrorMsg.ErrorMsg;
import Parse.*;
import Semant.*;
import Translate.*;
import Assem.*;
import Pentium.*;

/*
 * Java Oils and Touches a Tiger cOmpiler
 */
public class Jotto
{
    // stampa sintassi astratta?
    static boolean bPrintAbsyn = false;
    // stampa codice intermedio?
    static boolean bPrintIC = false;
    // stampa codice intermedio in forma canonica?
    static boolean bPrintCIC = false;
    // stampa codice instruction selection?
    static boolean bPrintIS = false;
    // stampa informazioni di debug? (nella fase di allocazione dei registri)
    static boolean bPrintDbgRA = false;
    // crea il file assembly?
    static boolean bMakeAsm = true;
    // dimensione dello stack in kb
    public static int stackSize = 256;

    static final String version = "v1.0";
    static final String[] headArr = {
    			"JOTTO - Java Oils and Touches a Tiger cOmpiler - "+version,
    			"sviluppato da Cristiano Tagliamonte",
                "              Cristiano Cenci",
                "              Giovanni Ottieri",
            };
    static final String header = headArr[0] + "\n" + headArr[1] + "\n" + headArr[2] + "\n" + headArr[3] + "\n";
    static final String usage = "Sintassi:\n"
                              + "jotto [FLAG[n]]... [OPT] [FILEIN] [FILEOUT] [FILEINFO]\n"
                              + "    FLAG sta per uno dei seguenti:\n"
                              + "        -a  :  mostra l'albero della sintassi astratta su stdout\n"
                              + "        -i  :  mostra codice intermedio su stdout\n"
                              + "        -c  :  mostra codice intermedio canonizzato stdout\n"
                              + "        -n  :  mostra instruction selection su stdout\n"
                              + "        -e  :  devia stderr su stdout\n"
                              + "        -d  :  genera file di debug per l'allocazione dei registri\n"
                              + "        -m  :  crea file assembly (abilitato per default)\n"
                              + "        usa il suffisso 'n' per dibabilitare il flag\n"
                              + "    OPT sta per uno dei seguenti:\n"
                              + "        -s DIM :  fissa la dimensione dello stack in kb (di base e' "+stackSize+"kb)\n"
                              + "        -h     :  show this help usage message\n"
                              + "    FILEIN e' il nome del file del sorgente tiger da compilare (di base e' stdin)\n"
                              + "    FILEOUT  e' il nome del file assembly generato\n"
                              + "    FILEINFO e' il nome del file su cui sono stampate i messaggi del compilatore (stdout per default)\n";

    static final String retry = "prova 'jotto -h' per mostrare l'aiuto.";

    // nome del file sorgente tiger
    static String srcFilename;
    // nome del file sorgente assembly da creare
    static String dstFilename;

    static PrintStream sysout=System.out;
    static PrintStream syserr=System.err;
    static PrintStream out=System.out;
    static PrintStream err=System.err;

    static void printexit(String s)
    {
        out.println(s);
        System.exit(1);
    }

    static int setFilename(String[] s, int i)
    {
        if(srcFilename != null)
        {
            out.println("Specificare un solo file sorgente Tiger!");
            printexit(retry);
            return i;
        }
        if(s != null)
        {
            if(!s[i].endsWith(".tig"))
            {
            	 out.println("Il file sorgente deve avere estensione '.tig'!");
                printexit(retry);
            }
            srcFilename = s[i];
            if(s.length>i+1)
            {
                int j=i+1;

                if(s[j].endsWith(".asm"))
                {
                    dstFilename = s[j];
                    j++;
                }
                else
                {
                    dstFilename = s[i].substring(0, s[i].length()-3) + "asm";
                }
                if(s.length>j)
                {
                    try
                    {
                        out=err=new PrintStream(new FileOutputStream(s[j]));
                        System.setOut(out);
                        System.setErr(err);
                    }
                    catch(Exception e)
                    {
                        e.printStackTrace();
                    }
                }
                i=j;
            }
        }
        else
            srcFilename= dstFilename = null;
        return i;
    }

    static void unknownpar(String s)
    {
        out.println("Parametro sconosciuto '"+s+"'.");
        printexit(retry);
    }

    static void parseargs(String[] args)
    {
        if(args.length == 0) {
            out.println(retry);
            return;
        }
        for(int i=0; i<args.length; i++) {
            if(args[i].charAt(0) != '-')
                i = setFilename(args, i);
            else {
                if(args[i].length() > 3) unknownpar(args[i]);
                else switch(args[i].charAt(1)) {
                    case 'a':
                        if(args[i].length() > 2) {
                            if(args[i].charAt(2) == 'n')
                                bPrintAbsyn = false;
                            else
                                unknownpar(args[i]);
                        }
                        else
                            bPrintAbsyn = true;
                        break;
                    case 'i':
                        if(args[i].length() > 2) {
                            if(args[i].charAt(2) == 'n')
                                bPrintIC = false;
                            else
                                unknownpar(args[i]);
                        }
                        else
                            bPrintIC = true;
                        break;
                    case 'c':
                        if(args[i].length() > 2) {
                            if(args[i].charAt(2) == 'n')
                                bPrintCIC = false;
                            else
                                unknownpar(args[i]);
                        }
                        else
                            bPrintCIC = true;
                        break;
                    case 'n':
                        if(args[i].length() > 2) {
                            if(args[i].charAt(2) == 'n')
                                bPrintIS = false;
                            else
                                unknownpar(args[i]);
                        }
                        else
                            bPrintIS = true;
                        break;
                    case 'e':
                        if(args[i].length() > 2) {
                            if(args[i].charAt(2) == 'n')
                                System.setErr(syserr);
                            else
                               unknownpar(args[i]);
                        }
                        else
                            System.setErr(out);
                        break;
                    case 'd':
                        if(args[i].length() > 2) {
                            if(args[i].charAt(2) == 'n')
                                bPrintDbgRA = false;
                            else
                               unknownpar(args[i]);
                        }
                        else
                            bPrintDbgRA = true;
                        break;
                    case 'm':
                        if(args[i].length() > 2) {
                        	if(args[i].charAt(2) == 'n')
                                bMakeAsm = false;
                            else
                                unknownpar(args[i]);
                        }
                        else
                            bMakeAsm = true;
                        break;
                    case 's':
                        try {
                            stackSize = Integer.parseInt(args[++i]);
                        } catch(Exception e) {
                            out.println("Specificare la dimensione dello stack!");
                            printexit(retry);
                        }
                        if(64 > stackSize || stackSize > 900) {
                            out.println("La dimensione dello stack deve essere tra 64kb e 900kb!");
                            printexit(retry);
                        }
                        break;
                    case 'h':
                        if(args[i].length() > 2)
                            unknownpar(args[i]);
                        else
                            printexit(usage);
                        break;
                    default:
                        unknownpar(args[i]);
                }
            }
        }
    }

    // ----> MAIN <----
    public static void main(String[] args)
    {
        out.println(header);
        parseargs(args);
        RegAlloc.Color.debug = bPrintDbgRA;

        // apre file .tig
        ErrorMsg errorMsg = new ErrorMsg(srcFilename);
        InputStream tigin = null;
        if(srcFilename == null)
            tigin = System.in;
        else {
            try {
                out.print("Caricamento file '"+srcFilename+"'... ");
                out.flush();
                tigin = new FileInputStream(srcFilename);
            } catch (FileNotFoundException e) {
                printexit("Errore: impossibile leggere il file '"+srcFilename+"'.");
            }
            out.println("fatto.\n");
        }
        
        out.println("---- Results");
        
        Parse parse=null;

        // analisi sintattica
        try {
            parse = new Parse(tigin, errorMsg);            
        }
        catch(Exception e) {
            out.println(e.getMessage());
            e.printStackTrace();
        }

        // analisi semantica (controllo dei tipi)
        // e generazione codice intermedio
        Semant semant = new Semant(errorMsg, new PFrame());
        try {
            semant.transProg((Absyn.Exp)parser.parseResult);
        }
        catch(SemanticError se) {
            errorMsg.error(se.getPos(), se.getMessage());
            System.exit(1);
        }
        if(errorMsg.semanticErrors == 0)
        {    
            out.println("\nAnalisi semantica completata con successo\n");
            if(bPrintAbsyn && parse!=null) parse.print(out);
        }
        else 
        {
            out.println("\nAnalisi semantica completata, "+errorMsg.semanticErrors+" errori\n");
            if(bPrintAbsyn && parse!=null) parse.print(out);
            System.exit(1);
        }

        // stampa codice intermedio
        Frag frags = semant.getFragments();
        if(frags == null) System.exit(1);
        if(bPrintIC ) semant.printIC(out);
        if(bPrintCIC) semant.printCanonIC(out);

        // instruction selection + register allocation
        String asm = "";
        String dataAsm = "";
        PFrame pframe = null;
        if(bPrintIS) out.println("\n\n---- Instruction Selection");
        for(Frag f=frags; f!=null; f=f.next) {
            if(f instanceof ProcFrag) {
                ProcFrag pfrag = (ProcFrag)f;
                pframe = (PFrame)pfrag.frame;
                InstrList instrs = new InstrList();
                for(Tree.StmList b=pfrag.cbody; b!=null; b=b.tail)
                    instrs.addInTail(pframe.codegen(b.head));
                if(bPrintIS) {
                    out.println("\n-- function '"+pframe.debugName+"'");
                    out.println("_"+pframe.name);
                    RegAlloc.Color.printInstrs(new Temp.DefaultMap(), out, instrs.first());
                }
                if(bMakeAsm) {
                	asm += "\n;-- function '"+pframe.debugName+"'\n";
                    asm += pframe.procEntryExit3(instrs.first()).code;
                }
            }
            else if(f instanceof DataFrag)
                dataAsm += ((DataFrag)f).data + " ;string: \"" + ((DataFrag)f).comment + "\"\n";
        }
        if(bPrintIS && dataAsm.length()>0) {
            out.println("\n-- data segment");
            out.println(dataAsm);
        }

        // compilazione completata, salva il file sorgente assembly
        if(bMakeAsm) {
        	try {
        		BufferedWriter asmout = new BufferedWriter(dstFilename == null ?
														   new OutputStreamWriter(out) :
                                                           new FileWriter(dstFilename));
        		// sezione del codice assembly
        		asmout.write(";----------------------------------------------------------\n");
                asmout.write(";--  file assembly generato da:\n");
                asmout.write(";--\n");
                for(int h=0; h<headArr.length; h++)
                    asmout.write(";--  " + headArr[h] + '\n');
                asmout.write(";--\n");
                asmout.write(";----------------------------------------------------------\n");
                asmout.write(asm);
        		asmout.write("\n");
        		// funzioni e dati esterni
        		java.util.Enumeration ext = Frame.Frame.extRoutine.keys();
        		while(ext.hasMoreElements())
        			asmout.write(".extern _" + ((String)ext.nextElement()) + "\n");
        		asmout.write(".extern _init\n");
        		asmout.write(".extern _checkStack\n");
        		asmout.write(".extern _Stack_Size\n");
        		asmout.write(".extern _Stack_Init\n");
        		asmout.write(".extern _Bound\n");
        		asmout.write(".extern _RV\n");
        		// sezione dati
        		asmout.write("\n.data\n" + dataAsm);
        		asmout.flush();
        		asmout.close();
        	} catch(Exception e) {
                if(dstFilename == null)
                    printexit("Errore: impossibile scrivere su stdout.");
                else
                	printexit("Errore: impossibile scrivere sul file '"+dstFilename+"'.");
        	}
        }
        if(System.out!=sysout)
            System.setOut(sysout);
        if(System.err!=syserr)
            System.setErr(syserr);
    }
}
