package RegAlloc;

import Assem.*;
import Temp.*;
import Graph.NodeList;
import Graph.Node;

import java.util.*;
import java.io.*;

//La classe Color effettua la colorazione del grafo mantenendo nella lista che ritorna spills
//i nodi che devono essere spillati (actual spill)

public class Color implements TempMap 
{

	//I formali passati al costruttore

 	TempMap init; 	
  TempNode regs; 	
 	InterferenceGraph ig;
 	//
 	PrintStream PS;
	static int count = 0;
	public static boolean debug = false;

	//Matrici di incidenza
	NodeList oldadjList[];
	
	//Stack nodi da assegnare
	NodeList actNodes = null;

	//Stack nodi da assegnare
	NodeList stack = null;

	//Lista pot. spill
	NodeList spilledNodes = null;

	//Lista nodi precolorati
	NodeList precolored = null;
	
	//Matrici di adiacenza
	boolean adjSet[][];

	//Matrici di incidenza
	NodeList adjList[];
	
	//Numero di colori
	int K;
	
	//Numero di nodi
	int nodes;
	
	//Mappa nodi - move
	Hashtable movelist = new Hashtable();

	//Alias
	Hashtable alias = new Hashtable();
	
	//Colori assegnati
	int color[];
	
	//Mappa Indice Colore Stringa
	String map[];
  
  public TempNode spills(){
  	return toTempList(spilledNodes);
  } 

  public String tempMap(Temp t){
  	if (in(regs,t))return init.tempMap(t);
  	if (color[key(node(t))] == -1) return "spill";
  	return map[color[key(node(t))]];
  }
 
  public Color(InterferenceGraph ig,TempMap initial,TempNode registers)
  {  	
		try{
  		    if(debug)
                PS = new PrintStream(new FileOutputStream("./debug/regalloc"+(count++)+".txt"));
        }  		
		catch(FileNotFoundException ff){
			//System.out.println("Errore in creazione file di log per allocazione registri");
		}
		/**/
		//PS = System.out;

  	init = initial;
  	this.ig = ig;
  	regs = registers;
  	K = num(regs);
  	if(debug)
        PS.println("K = " + K);
  	nodes = num(ig.nodes());
  	if(debug)
        PS.println("nodi = " + nodes);	
	go();
  }

  void init(){
  	//bisogna inizializzare le varie strutture dati sulla base dell'interference Graph
  	
  	//Crea le strutture Ad Array
  	map = new String[K];
  	color = new int[nodes];
  	adjSet = new boolean[nodes][nodes];
  	adjList = new NodeList[nodes];
  	oldadjList = new NodeList[nodes];
		//Tutti i nodi del grafo
		initoldList();
  	NodeList nods = ig.nodes();
  	//Inizializza Matrici e Liste di adiacenza
  	while(nods!=null){  		
  	  Node g = nods.head;
  	  color[key(g)]=-1;
 	  reAdd(g);
  	  nods = nods.tail;
  	}
  	//Inizializza i nodi precolored
  	precolored = toNodeList(regs);
  	TempNode preCol = regs;
  	int jj=0;
  	while(preCol!=null){
  		map[jj] = init.tempMap(preCol.head);  		
  		if(node(preCol.head)!=null)color[key(node(preCol.head))] = jj;
  		jj++;
  		preCol = preCol.tail;
  	}
  }

  void go(){
  	init();
  	if(debug)printInfo();
  	if(debug)printMatrix();
  	if(debug)printPreCol();
  	if(debug)PS.println("Traccia Algoritmo");
  	simplify();
  	//Ho solo precolorati non nello stack
  	AssignColors();
  	if(debug)PS.println("Fine Traccia Algoritmo");
  	if(debug)printColors();
  	if(debug)printSpilled();
  }

  void simplify(){
  	//cerca un nodo inseribile nello stack
  	Node s = null;
  	NodeList nl = actNodes;
  	while(nl!=null){
  		if ((degree(nl.head) < K) && (!precolored(nl.head))){
  			s=nl.head;
  			break;
  		}
  		nl = nl.tail;
  	}
  	if (s != null){
  		//rimuovi s dal grafo decrementa i gradi degli adiacenti
  		if(debug)PS.println("");
  		if(debug)PS.println("Pushed & Removed" + temp(s));
  		push(s);
  		remove(s);
   		simplify();
  	}
  	else{
  		SelectSpill();
  	}
  }

  void SelectSpill(){
  	Node s = null;
  	NodeList nl = actNodes;
  	//Euristica : il primo che trovo non precolorato
  	while(nl!=null){
  		if (!precolored(nl.head)){
  			s = nl.head;
  		}
  		nl = nl.tail;
  	} 
  	if (s==null)return;
  	else{
  		if(debug)PS.println("");
  		if(debug)PS.println("Pushed After Pot.Spill " + temp(s));
  		push(s);
  		remove(s);
  		simplify();
  	}
  }
  
  void AssignColors(){
  	boolean canAss = false;
  	while(stack!=null){
  		Node n = pop();
  		if(debug)PS.println("");
  		if(debug)PS.println("Popped & ReAdded" + temp(n));
  		reAdd(n);
  		canAss = false;
  		int i = 0;
  		while ((!canAss)&&(i<K)){
  			if(debug)
                PS.println("Assignable " + temp(n) + " color " + i + " " + (assignable(n,i)));
  			canAss = assignable(n,i);  		  
  			if (!canAss)i++;
  		}
  		if(!canAss){
  			spilledNodes = ins(spilledNodes,n);
  		}
  		else {
  			color[key(n)] = i;
  			if(debug)PS.println("Node " + temp(n) + " can assign color " + i + ":" + map[i]);
  		}
  	}
  }

  boolean assignable(Node k, int i){
  	boolean app = true;
  	NodeList adjs = adjList[key(k)];
  	while(adjs!=null){
  		if (color[key(adjs.head)] == i) return false;
  		adjs = adjs.tail;
  	}
  	return app;
  }
  

  // Funzioni sulle liste
	
	//NodeList  

  NodeList del(NodeList l , Node a){
  	if (l==null)return null;
  	if (a==l.head) return l.tail;
  		else return new NodeList(l.head,del(l.tail,a));
  }

  NodeList ins(NodeList l , Node a){
  	if (in(l,a)) return l;
  	return new NodeList(a,l);
  }
  
  boolean in(NodeList l , Node a){
    if(l==null)return false;
    if(l.head == a ) return true;
    	else return in(l.tail,a);    
  }

  boolean in(TempNode l , Temp a){
    if(l==null)return false;
    if(l.head == a ) return true;
    	else return in(l.tail,a);    
  }

  int num(NodeList t){
  	if (t == null) return 0;
  	else return 1 + num(t.tail);
  }

  NodeList toNodeList(TempNode n){
  	if (n == null)return null;
  	else return new NodeList(ig.tnode(n.head),toNodeList(n.tail));  
  }

  TempNode toTempList(NodeList n){
  	if (n == null)return null;
  	else return new TempNode(ig.gtemp(n.head),toTempList(n.tail));  
  }

  int num(TempNode t){
  	if (t == null) return 0;
  	else return 1 + num(t.tail);
  }
  
  //Degub e Stampa delle Strutture Dati

	void printMatrix(){
        if(debug) {
		PS.println("");
		PS.println("Matrice di Adiacenza");
    for (int i=0; i <nodes ; i++){
	  	for (int j=0; j <nodes ; j++){
	  		if (adjSet[i][j]) PS.print("1");
	  		else PS.print("0");
  		}  	
	  	PS.println("");
  	}
        }
	}
	
	void printSpilled(){
        if(debug) {
		if (spilledNodes!=null){
			PS.println("");
			PS.println("Temp da Spillare :");
			TempNode h = toTempList(spilledNodes);	
			while(h!=null){
				Temp gg = h.head;
				PS.print(key(node(gg)) + "/" + gg);
				PS.print(" ");
				h=h.tail;
			}
		}
        }
	}

	void printInfo(){
        if(debug) {
		PS.println("");
		PS.println("Info sui Nodi del Grafo:");
		NodeList h = actNodes;
		while(h!=null){
			Node g = h.head;
			PS.println("");
			PS.println("Nodo" + key(g) + "/" + temp(g));
			PS.println("Adiacenti");
			NodeList adjs = adjList[key(g)];
			while(adjs!=null){
				Node gg = adjs.head;
				PS.print(key(gg) + "/" + temp(gg));
				PS.print(" ");
				adjs = adjs.tail;
			}
			h=h.tail;
		}	
        }
    }	

	void printdegree(){
        if(debug) {
		PS.println("");
		PS.println("Gradi dei Nodi :");
		NodeList t = actNodes;
		while(t!=null){
			PS.println(ig.gtemp(t.head) + " " + degree(t.head));
			t=t.tail;
		}		
        }
	}

	void printColors(){
        if(debug) {
		PS.println("");
		PS.println("Colorazione :");
		NodeList t = actNodes;
		while(t!=null){
			PS.println(temp(t.head) + " " + tempMap(temp(t.head)));
			t=t.tail;
		}		
        }
	}
	
	void printPreCol(){
        if(debug) {
		PS.println("");
		PS.println("Precolorati :");
		TempNode prec = regs;
		while(prec!=null){
			PS.println(prec.head + " has Color " + init.tempMap(prec.head) + " in Graph ? " + (node(prec.head)!=null));
			prec = prec.tail;
		}	
        }
	}
	
    public static void printInstrs(TempMap t ,String file,InstrNode s){
        PrintStream ps = null;
        try{
            ps = new PrintStream(new FileOutputStream(file));
        }
        catch(FileNotFoundException ff){
            return;
        }
        printInstrs(t, ps, s);
    }

    public static void printInstrs(TempMap t, PrintStream ps,InstrNode s){
		InstrNode h = s;
		while (h!=null){
			String app = h.head.format(t);
			String app1 = "";
			for (int i=0;i<app.length();i++){
				if ((app.charAt(i)!='\r') && (app.charAt(i)!='\n')){
				  app1 = app1 + app.charAt(i);	
				}
			}
			app = app.substring(0,app.length()-1);
			ps.println(app);	
			h=h.tail;
		}			
	}
	
 	//Utilit
	boolean precolored(Node n){
		return in(precolored,n);
	}

  void reAdd(Node n){
  	NodeList adjs = oldadjList[key(n)];
  	while(adjs!=null){
  		insadj(n,adjs.head);
  		insadj(adjs.head,n);
  		adjSet[key(n)][key(adjs.head)] = true;
  		adjSet[key(adjs.head)][key(n)] = true;
  		adjs = adjs.tail;  		
  	}		
		actNodes = ins(actNodes,n);
  }

	void remove(Node n){
		NodeList adjs = adjList[key(n)];
		while (adjs != null){
			//cancella n dalle liste dei suoi adiacenti
			adjList[key(n)] = del(adjList[key(n)],adjs.head);
			adjList[key(adjs.head)] = del(adjList[key(adjs.head)],n);
			adjSet[key(n)][key(adjs.head)] = false;
			adjSet[key(adjs.head)][key(n)] = false;
		  adjs = adjs.tail;
		}
		actNodes = del(actNodes,n);
	}

  int key(Node g){
  	return (new Integer(g.toString())).intValue();
  }
  
  int degree(Node g){
  	return num(adjList[key(g)]);  
  }
    
  void insadj(Node g, Node h){
  	adjList[key(g)] = ins(adjList[key(g)],h);
  }
  
  void push(Node g){
  	stack = new NodeList(g,stack);
  }
  
  Node pop(){
  	Node app = stack.head;
  	stack = stack.tail;
  	return app;
  }
  Temp temp(Node g){
  	return ig.gtemp(g);
  }
  Node node(Temp t){
  	return ig.tnode(t);
  }
	void initoldList(){
		NodeList nods = ig.nodes();
  	while(nods!=null){
  	  Node g = nods.head;
  	  NodeList adjs = g.adj();
  	  while (adjs != null){
  	   Node aj = adjs.head;
  	   oldadjList[key(g)] = ins(oldadjList[key(g)],aj);
  	   oldadjList[key(aj)] = ins(oldadjList[key(aj)],g);
  	   adjs = adjs.tail;
  	  }
  	  nods = nods.tail;
  	}
	}

}