package RegAlloc;

import FlowGraph.*;
import Temp.*;
import Assem.*;

public class RegAlloc implements TempMap
{
  public InstrNode instrs;
  public Frame.Frame frame;
  public static int count = 0;
  int actcount = 0;
  Color col;

  public RegAlloc(Frame.Frame f, Assem.InstrNode il)
  {
  	frame = f;
  	instrs = il;
  	actcount = count++;
 	if(Color.debug) 
       Color.printInstrs(new DefaultMap(),"debug\\beforecol"+actcount+".txt",il);
  	main();
  }

  public String tempMap(Temp t)
  {
  	if (col!=null)return col.tempMap(t);
  	else return ((Pentium.PFrame)frame).tempMap(t);
  }
  
  void main()
  {
  	//f  la mappa dei nodi che vanno precolorati e la cui implementazione va in Pentium.PFrame
 	  Color c = new Color(new Liveness(new AssemFlowGraph(instrs)),(Pentium.PFrame)frame,((Pentium.PFrame)frame).registers());
      if(c.spills() != null) {
 	  	// Spilla i nodi da spillare e ricomincia
        rewriteProgram(c.spills());
 	  	main();
 	  }
 	  //Se non ci sono nodi da spillare la colorazione c  quella giusta.
 	  else {
 	  	col = c;
 	  	if(Color.debug)Color.printInstrs(col,"debug\\aftercol"+actcount+".txt",instrs);
 	  	optimize();
 	  	if(Color.debug)Color.printInstrs(col,"debug\\afteropt"+actcount+".txt",instrs);
 	  }
  }

  void rewriteProgram(TempNode lst)
  {
  	if (lst == null) return;
  	spill(lst.head);
  	rewriteProgram(lst.tail);
  }

  void spill(Temp v)
  {
		//Frame.Access na = frame.allocLocal(true);	
 		//Alloca lo spazio nel frame mi serve l'offset : devo sapere quante locali sono
 		//state allocate nel frame
	int off = newAlloc();
  	
  	InstrNode il = instrs;
  	InstrNode prec = null;
  	InstrNode newL = null; 
  	
  	while(il!=null) {
  		TempNode use =il.head.use();
  		TempNode def =il.head.def();
  		boolean cur = false;

  		//Se Utilizza
  		if (in(use,v)) {
  			Instr In = 
  				new OPER("movl "+off+"(%`s0),%`d0\n",
  					new TempNode(v,null),
  					new TempNode(frame.FP(),null)
  				);
  			//inserisci l'istruzione creata prima di quella corrente
  			newL = new InstrNode(In,newL);
  			newL = new InstrNode(il.head,newL);
  			cur  = true;
  		}

  		//Se definisce
  		if (in(def,v)){
  			Instr In = 
  				new OPER("movl %`s1,"+off+"(%`s0)\n",
  					null,
  					new TempNode(frame.FP(),new TempNode(v,null))
  				);
  			//inserisci l'istruzione creata dopo di quella corrente
  			if (!cur) 
                newL = new InstrNode(il.head,newL);
  			newL = new InstrNode(In,newL);
  			cur  = true;
  		}
  		if(!cur)
            newL = new InstrNode(il.head,newL);
  		il = il.tail;
  	}
 		instrs = reverse(newL);
  }

  //Elimina le move ridondati dopo aver colorato ci si rende conto di queste move

  void optimize(){
  	InstrNode il = instrs;
  	InstrNode newL = null; 
  	boolean added;
  	while(il!=null){
  		Instr i = il.head;
  		added = false;
  		if (i instanceof Assem.MOVE){
  			if ((num(i.def()) == 1) && (num(i.use()) == 1)){
  				if( (col.tempMap(i.def().head)).compareTo(col.tempMap(i.use().head)) == 0){
  					added = true;
  				}
  			}
  		}
  		if(!added)newL = new InstrNode(i,newL);
  		il = il.tail;  	
  	}
  	instrs = reverse(newL);
  }

	String signed(int i){
		if (i>0) return ("+"+i);
		else return (""+i);
	}

  boolean in(TempNode l , Temp a){
  	if(l==null)return false;
   		if(l.head == a ) return true;
    	else return in(l.tail,a);    
  }

  int newAlloc(){
  	Pentium.InFrame acc = (Pentium.InFrame)frame.allocLocal(true);
  	return acc.offset;
  }
  
  InstrNode reverse(InstrNode il){
  	InstrNode j = null;
  	while(il != null){
  		j = new InstrNode(il.head,j);
  		il = il.tail;
  	}
  	return j;
  }

  int num(TempNode t){
  	if (t == null) return 0;
  	else return 1 + num(t.tail);
  }
  
}